# Modified version of https://github.com/mubeta06/python/blob/master/signal_processing/sp/gold.py
# Lots of weird dependencies and unneeded functions were removed and placed directly here

import numpy
from numpy import fft
import pylab
from PIL import Image
import matplotlib.pyplot as plt
from scipy import misc
from scipy.signal import correlate
import scipy.io.wavfile

preferred_pairs = {5:[[2],[1,2,3]], 6:[[5],[1,4,5]], 7:[[4],[4,5,6]],
                        8:[[1,2,3,6,7],[1,2,7]], 9:[[5],[3,5,6]],
                        10:[[2,5,9],[3,4,6,8,9]], 11:[[9],[3,6,9]]}
# nbits can only be 5, 6, 7, 8, 9, 10, 11

def bw_load_image_as_array(file_name):
    arr = numpy.array(Image.open(file_name).convert(mode='L'))
    bw_arr = []
    for x in arr:
        bw_arr_row  = []
        for y in x:
            if (y > 128):
                bw_arr_row += [1]
            else:
                bw_arr_row += [0]
        bw_arr += [bw_arr_row]
    return numpy.array(bw_arr)

def test_vectorize_and_adjust(func, test_image):
    correct = numpy.array([1 if x > 0 else -1 for x in test_image.flatten()])
    student = func(test_image)
    if (not numpy.array_equal(correct, student)):
        print("Incorrect! The arrays do not match.")
    else:
        print("Correct!")
    return

def load_image_as_array(file_name):
    return numpy.array(Image.open(file_name).convert(mode='L'))

def lfsr(taps, buf):
    nbits = len(buf)
    sr = numpy.array(buf, dtype='bool')
    out = []
    for i in range((2**nbits)-1):
        feedback = sr[0]
        out.append(feedback)
        for t in taps:
            feedback ^= sr[t]
        sr = numpy.roll(sr, -1)
        sr[-1] = feedback
    return out

def gen_gold(seq1, seq2):
    """Function to produce a gold sequence based on two input preferred pair
    Maximal Length Sequences
    """
    gold = [seq1, seq2]
    for shift in range(len(seq1)):
        gold.append(numpy.logical_xor(seq1, numpy.roll(seq2, -shift)))
    return gold

def gold(n):
    """Generate a set of 2^n +1 Gold Codes
    """
    n = int(n)
    if not n in preferred_pairs:
        raise ss.Error('preferred pairs for %s bits unknown' % str(n))
    seed = list(numpy.ones(n))
    seq1 = lfsr(preferred_pairs[n][0], seed)
    seq2 = lfsr(preferred_pairs[n][1], seed)
    return [numpy.where(x, 1, -1) for x in gen_gold(seq1, seq2)]

def get_gold_codes(n):
    return numpy.array(gold(10))[:n]

def correct_correlation(signal_one, signal_two):
    return vector_norm(numpy.correlate(signal_one, signal_two, 'same'))

def test_correlation(cross_correlation, signal_one, signal_two):
#    result_lib = numpy.convolve(signal_one, signal_two[::-1])
    # result_lib = numpy.array([numpy.correlate(signal_one, numpy.roll(signal_two, k)) for k in range(len(signal_two))])
    result_lib = correct_correlation(signal_one, signal_two)
    result_stu = cross_correlation(signal_one, signal_two)
    return result_lib, result_stu

def test_correlation_plot(signal1, signal2, lib_result, your_result):
    # Plot the output
    fig = plt.figure(figsize=(8,3))
    ax = plt.subplot(111)
    str_corr='Correct Answer (length='+str(len(lib_result))+')'
    str_your='Your Answer (length='+str(len(your_result))+')'

    ax.plot([x-len(signal2)+1 for x in range(len(lib_result))], lib_result, 'k', label=str_corr, lw=1.5)
    ax.plot([x-len(signal2)+1 for x in range(len(your_result))], your_result, '--r', label=str_your, lw = 3)
    ax.set_title("Cross correlation of:\n%s\n%s"%(str(signal1), str(signal2)))
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))

def test_single_channel_bw_gold_code_encoding(func, gold_code):
    adj_bit_stream = [1, -1, 1, 1, -1]
    print("Test adjusted bit stream:\n{}\n".format(adj_bit_stream))
    print("=====================================\n")
    encoded_bit_stream = []
    for bit in adj_bit_stream:
        encoded_bit_stream += [bit * g for g in gold_code]
    correct =  numpy.array(encoded_bit_stream)

    student_answer = func(adj_bit_stream, gold_code)
    if (len(correct) != len(student_answer)):
        print("Incorrect! The lengths do not match: {} (your answer) and {} (correct answer)".format(
            len(student_answer), len(correct)))
    elif (not numpy.array_equal(correct, student_answer)):
        print("Incorrect! The values do not match.")
    elif (not isinstance(student_answer, numpy.ndarray)):
        print("Invalid Type! Please cast as np.array(YOUR_LIST)")
    else:
        print("Correct!")
    return

def save_clip(signal, file_name, rate=44100, scale=20000):
    """Saves a signal as a 16 bit PCM WAV file
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.io.wavfile.write.html
    """
    signal_scaled_to_write = (signal * (scale // 2 // max(signal))).astype(numpy.int16)
    scipy.io.wavfile.write(file_name, rate, signal_scaled_to_write)
    print("Saved to file {}".format(file_name))
    return signal_scaled_to_write

def load_clip(file_name):
    rate, recording = scipy.io.wavfile.read(file_name)
    if (rate != 44100):
        print("ERROR: Recording should be at 44.1kHz sample. Got {}.\n".format(rate))
        return None
    elif (len(recording) == 2):
        return recording[0]
    else:
        return recording

def simple_gold_code_enc_fig():
    fig = numpy.array([1,1,1,-1,1,1,1,1,-1,-1,-1,-1,-1,1,-1,-1])
    sig = [0, 1, 0, 1, 1, 0]
    encode = []
    for x in fig:
        encode += [x * s for s in sig]
    print(encode)

    plt.figure(figsize=(15,5))
    plt.step(range(len(encode)), encode)
    plt.title("First 100 elements of the gold code encoded image")
    plt.show()

def simple_gold_code_enc_fig_noise():
    fig = numpy.array([1,1,1,-1,1,1,1,1,-1,-1,-1,-1,-1,1,-1,-1])
    sig = [0, 1, 0, 1, 1, 0]
    encode = []
    for x in fig:
        encode += [(x * s * 100)/1.0 + numpy.random.randint(-5, 5)  for s in sig]
    encode = [e / 100.0 for e  in encode]
    print(numpy.array(encode))

    plt.figure(figsize=(15,5))
    plt.step(range(len(encode)), encode)
    plt.title("First 100 elements of the gold code encoded image")
    plt.show()

def simple_gold_code_enc_fig_noisy_xcorr():
    corr = cross_correlation(sig, encode)

    plt.figure(figsize=(15,5))
    plt.step(range(len(corr)), corr)
    plt.title("First 100 elements of the gold code encoded image")
    plt.show()

def vector_norm(vec):
    return vec / abs(max(vec.min(), vec.max(), key=abs))

def normalize_slices(correlations, slice_size=500000):
    norm_cor = []
    for i in range((len(correlations) // slice_size) + 1):
        norm_cor += vector_norm(correlations[i*slice_size : min(i*slice_size+slice_size, len(correlations))]).tolist()
    return norm_cor

def best_attempt_threshold(cross_corr_vec, x, y, threshold=0.9, limit=25, debug=False):
    binned_size = 0
    count = 0
    target = x * y
    while (binned_size != target and count < limit):
        count += 1
        binned_size = 0
        for c in cross_corr_vec:
            if (abs(c) > threshold):
                binned_size += 1
        if (binned_size > target):
            threshold += 0.005
        elif (binned_size < target):
            threshold -= 0.005
    if (count == limit and debug):
        print("Could not converge - Threshold: {} - Binned size: {} - Target: {}".format(
            threshold, binned_size, target))
    elif (debug):
        print("Converged - Threshold: {} - Binned size: {} - Target: {}".format(
            threshold, binned_size, target))
    return threshold

def get_real_recording(file_name='real_recording.wav'):
    return scipy.io.wavfile.read(file_name)[1]

def gen_ideal_recording(encode, vec_adj, gold_code, file_name='sprites/pastry_feline.png'):
    """Returns an ideal recording (the gold code encoded bit stream) of the test image

    :param encode: Black and White encoding function from the notebook
    :type  encode: <Function>

    :param vec_adj: Vectorize and Adjiust function from the notebook
    :type  vec_adj: <Function>

    :param gold_code: Gold code used to encode
    :type  gold_code: Numpy array

    :param file_name: File name of the image (Defaults to pastry feline)
    :type  file_name: <String>
    """
    return encode(vec_adj(bw_load_image_as_array(file_name)), gold_code)

def plot_ideal_encoding(cross_corr_decode, ideal_recording_cross_corr, threshold=0, X=32, Y=20):
    ideal_decoded_vector = cross_corr_decode(ideal_recording_cross_corr, threshold=0, X=X, Y=Y)
    try:
        ideal_decoded_image = numpy.reshape(ideal_decoded_vector, (Y, X))
    except ValueError:
        print("=== Failed Ideal Image Decoding - adjust your threshold ===\n")
        ideal_decoded_image = bw_load_image_as_array('sprites/error.png')
    plt.figure(figsize=(20,5))
    plt.subplot(1, 2, 1)
    plt.plot(ideal_recording_cross_corr, 'r')
    plt.plot([threshold] * len(ideal_recording_cross_corr), 'g')
    plt.plot([-threshold] * len(ideal_recording_cross_corr), 'g')
    plt.title("Ideal Recording (Cross Correlation Peaks)")
    plt.subplot(1, 2, 2)
    plt.imshow(ideal_decoded_image, cmap = 'gray', interpolation = 'nearest')
    plt.title("Ideal Reconstructed Image")
    plt.tight_layout()
    plt.show()

def ideal_real_comparison(cross_corr_decode, ideal_recording_cross_corr, real_recording_cross_corr, threshold, X=32, Y=20):
    ideal_decoded_vector = cross_corr_decode(ideal_recording_cross_corr, threshold=threshold, X=X, Y=Y)
    real_decoded_vector  = cross_corr_decode(real_recording_cross_corr, threshold=threshold, X=X, Y=Y)

    print("Length of ideal decoded vector: {}\n".format(len(ideal_decoded_vector)))
    print("Length of real decoded vector:  {}\n".format(len(real_decoded_vector)))

    try:
        real_decoded_image = numpy.reshape(real_decoded_vector, (Y, X))
    except ValueError:
        print("=== Failed Real Image Decoding - adjust your threshold ===\n")
        real_decoded_image = bw_load_image_as_array('sprites/error.png')

    try:
        ideal_decoded_image = numpy.reshape(ideal_decoded_vector, (Y, X))
    except ValueError:
        print("=== Failed Ideal Image Decoding - adjust your threshold ===\n")
        ideal_decoded_image = bw_load_image_as_array('sprites/error.png')

    plt.figure(figsize=(20,5))
    plt.subplot(1, 2, 1)
    plt.plot(ideal_recording_cross_corr, 'r')
    plt.plot([threshold] * len(ideal_recording_cross_corr), 'g')
    plt.plot([-threshold] * len(ideal_recording_cross_corr), 'g')
    plt.title("Ideal Recording (Cross Correlation Peaks)")
    plt.subplot(1, 2, 2)
    plt.imshow(ideal_decoded_image, cmap = 'gray', interpolation = 'nearest')
    plt.title("Ideal Reconstructed Image")
    plt.tight_layout()
    plt.show()

    plt.figure(figsize=(20,5))
    plt.subplot(1, 2, 1)
    plt.plot(real_recording_cross_corr, 'b')
    plt.plot([threshold] * len(real_recording_cross_corr), 'g')
    plt.plot([-threshold] * len(real_recording_cross_corr), 'g')
    plt.title("Real Recording (Cross Correlation Peaks)")
    plt.subplot(1, 2, 2)
    plt.imshow(real_decoded_image, cmap = 'gray', interpolation = 'nearest')
    plt.title("Real Reconstructed Image")
    plt.tight_layout()
    plt.show()

def reshape_and_display(decoded_vector, X=32, Y=20):
    reconstructed_image = decoded_vector.reshape(Y,X)
    plt.imshow(reconstructed_image, cmap='gray', interpolation='nearest')
    plt.title("Reconstructed Image")
    plt.show()

def simulate_noise(vector):
    std = 2
    return numpy.multiply(vector, numpy.add(numpy.array([1] * len(vector)), numpy.random.normal(0,std,len(vector))))

def sprites_test(save_gold_code_encoded_image, decode_gold_code_encoded_recording, gold_code, cross_correlation, cross_corr_decode_bin):
    sprite_tuples = [('sprites/sprite_fairy.png', 'encoded_fairy.wav'),
                     ('sprites/sprite_monster.png', 'encoded_monster.wav'),
                     ('sprites/sprite_bird.png', 'encoded_bird.wav'),
                     ('sprites/sprite_fish.png', 'encoded_fish.wav')]
    for t in sprite_tuples:
        image_name, wav_name = t
        ideal_signal = save_gold_code_encoded_image(image_name, wav_name, gold_code)
        real_signal  = simulate_noise(decode_gold_code_encoded_recording(wav_name, gold_code, X=16, Y=16))
        ideal_recording_cross_corr = cross_correlation(ideal_signal, gold_code)
        real_recording_cross_corr  = cross_correlation(real_signal, gold_code)
        ideal_real_comparison(cross_corr_decode_bin, ideal_recording_cross_corr, real_recording_cross_corr, 0, X=16, Y=16)

print("Imported Gold Codes Helper Functions!")
